@@ -1,4 +1,123 @@
-"""Module with core logic to interact with OpenAPI APIs."""
+# region: docstring
+"""
+# OpenApiLibCore for Robot Framework®
+
+The OpenApiLibCore library is a utility library that is meant to simplify creation
+of other Robot Framework libraries for API testing based on the information in
+an OpenAPI document (also known as Swagger document).
+This document explains how to use the OpenApiLibCore library.
+
+For more information about Robot Framework®, see http://robotframework.org.
+
+---
+
+> Note: OpenApiLibCore is still being developed so there are currently
+restrictions / limitations that you may encounter when using this library to run
+tests against an API. See [Limitations](#limitations) for details.
+
+---
+
+## Installation
+
+If you already have Python >= 3.8 with pip installed, you can simply run:
+
+`pip install --upgrade robotframework-openapi-libcore`
+
+---
+
+## OpenAPI (aka Swagger)
+
+The OpenAPI Specification (OAS) defines a standard, language-agnostic interface
+to RESTful APIs, see https://swagger.io/specification/
+
+The OpenApiLibCore implements a number of Robot Framework keywords that make it
+easy to interact with an OpenAPI implementation by using the information in the
+openapi document (Swagger file), for examply by automatic generation of valid values
+for requests based on the schema information in the document.
+
+> Note: OpenApiLibCore is designed for APIs based on the OAS v3
+The library has not been tested for APIs based on the OAS v2.
+
+---
+
+## Getting started
+
+Before trying to use the keywords exposed by OpenApiLibCore on the target API
+it's recommended to first ensure that the openapi document for the API is valid
+under the OpenAPI Specification.
+
+This can be done using the command line interface of a package that is installed as
+a prerequisite for OpenApiLibCore.
+Both a local openapi.json or openapi.yaml file or one hosted by the API server
+can be checked using the `prance validate <reference_to_file>` shell command:
+
+```shell
+prance validate http://localhost:8000/openapi.json
+Processing "http://localhost:8000/openapi.json"...
+ -> Resolving external references.
+Validates OK as OpenAPI 3.0.2!
+
+prance validate /tests/files/petstore_openapi.yaml
+Processing "/tests/files/petstore_openapi.yaml"...
+ -> Resolving external references.
+Validates OK as OpenAPI 3.0.2!
+```
+
+You'll have to change the url or file reference to the location of the openapi
+document for your API.
+
+If the openapi document passes this validation, the next step is trying to do a test
+run with a minimal test suite.
+The example below can be used, with `source`, `origin` and 'endpoint' altered to
+fit your situation.
+
+``` robotframework
+*** Settings ***
+Library            OpenApiLibCore
+...                    source=http://localhost:8000/openapi.json
+...                    origin=http://localhost:8000
+
+*** Test Cases ***
+Getting Started
+    ${url}=    Get Valid Url    endpoint=/employees/{employee_id}   method=get
+
+```
+
+Running the above suite for the first time may result in an error / failed test.
+You should look at the Robot Framework `log.html` to determine the reasons
+for the failing tests.
+Depending on the reasons for the failures, different solutions are possible.
+
+Details about the OpenApiLibCore library parameters and keywords that you may need can be found
+[here](https://marketsquare.github.io/robotframework-openapi-libcore/openapi_libcore.html).
+
+The OpenApiLibCore also support handling of relations between resources within the scope
+of the API being validated as well as handling dependencies on resources outside the
+scope of the API. In addition there is support for handling restrictions on the values
+of parameters and properties.
+
+Details about the `mappings_path` variable usage can be found
+[here](https://marketsquare.github.io/robotframework-openapi-libcore/advanced_use.html).
+
+---
+
+## Limitations
+
+There are currently a number of limitations to supported API structures, supported
+data types and properties. The following list details the most important ones:
+- Only JSON request and response bodies are currently supported.
+- The unique identifier for a resource as used in the `paths` section of the
+    openapi document is expected to be the `id` property on a resource of that type.
+- Limited support for query strings and headers.
+- Limited support for authentication
+    - `username` and `password` can be passed as parameters to use Basic Authentication
+    - A [requests AuthBase instance](https://docs.python-requests.org/en/latest/api/#authentication)
+        can be passed and it will be used as provided.
+    - No support for per-endpoint authorization levels (just simple 401 validation).
+- byte, binary, date, date-time string formats not supported yet.
+
+"""
+# endregion
 
 import json as _json
 import sys
@@ -158,7 +277,12 @@ class RequestData:
 
 @library(scope="TEST SUITE", doc_format="ROBOT")
 class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
-    """Main class providing the keywords and core logic to interact with an OpenAPI."""
+    """
+    Main class providing the keywords and core logic to interact with an OpenAPI server.
+
+    Visit the [https://github.com/MarketSquare/robotframework-openapi-libcore | library page]
+    for an introduction.
+    """
 
     def __init__(  # pylint: disable=too-many-arguments, too-many-locals
         self,
@@ -172,6 +296,40 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
         auth: Optional[AuthBase] = None,
         invalid_property_default_response: int = 422,
     ) -> None:
+        # region: docstring
+        """
+        === source ===
+        An absolute path to an openapi.json or openapi.yaml file or an url to such a file.
+
+        === origin ===
+        The server (and port) of the target server. E.g. ``https://localhost:7000``
+
+        === base_path ===
+        The routing between ``origin`` and the endpoints as found in the ``paths`` in the
+        openapi document. E.g. ``/petshop/v2``.
+
+        === mappings_path ===
+        See [https://marketsquare.github.io/robotframework-openapi-libcore/advanced_use.html | here].
+
+        === username ===
+        The username to be used for Basic Authentication.
+
+        === password ===
+        The password to be used for Basic Authentication.
+
+        === security_token ===
+        The token to be used for token based security using the ``Authorization`` header.
+
+        === auth ===
+        A [https://docs.python-requests.org/en/latest/api/#authentication | requests AuthBase instance]
+        to be used for authentication instead of the ``username`` and ``password``.
+
+        === invalid_property_default_response ===
+        The default response code for requests with a JSON body that does not comply with
+        the schema. Example: a value outside the specified range or a string value for a
+        property defined as integer in the schema.
+        """
+        # endregion
         try:
             parser = ResolvingParser(source, backend="openapi-spec-validator")
         except ResolutionError as exception:
@@ -218,6 +376,7 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
 
     @property
     def openapi_spec(self):
+        """Return a deepcopy of the parsed openapi document."""
         # protect the parsed openapi spec from being mutated by reference
         return deepcopy(self._openapi_spec)
 
@@ -231,7 +390,7 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
 
         > Note: if valid ids cannot be retrieved within the scope of the API, the
         `PathPropertiesConstraint` Relation can be used. More information can be found
-        [here](https://marketsquare.github.io/robotframework-openapidriver/advanced_use.html).
+        [https://marketsquare.github.io/robotframework-openapi-libcore/advanced_use.html | here].
         """
         method = method.lower()
         try:
@@ -690,15 +849,9 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
         Returns a version of `params, headers` as present on `request_data` that has
         been modified to cause the provided `status_code`.
         """
-        if not request_data.params and not request_data.headers:
-            raise ValueError("No params or headers to invalidate.")
         if not request_data.parameters:
-            raise ValueError(
-                "Could not invalidate parameters: parameters list was empty."
-            )
-        # ensure we're not modifying mutable properties
-        params = deepcopy(request_data.params)
-        headers = deepcopy(request_data.headers)
+            raise ValueError("No params or headers to invalidate.")
+        # ensure the status_code can be triggered
         relations = request_data.dto.get_parameter_relations_for_error_code(status_code)
         relations_for_status_code = [
             r
@@ -706,44 +859,59 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
             if isinstance(r, PropertyValueConstraint) and r.error_code == status_code
         ]
         relation_property_names = {r.property_name for r in relations_for_status_code}
-        parameter_names = set(params.keys()).union(set(headers.keys()))
-        if parameter_names.isdisjoint(relation_property_names):
+        if not relation_property_names:
             if status_code != self.invalid_property_default_response:
                 raise ValueError(
                     f"No relations to cause status_code {status_code} found."
                 )
-            parameter_to_invalidate = choice(tuple(parameter_names))
-            try:
-                [parameter_data] = [
-                    data
-                    for data in request_data.parameters
-                    if data["name"] == parameter_to_invalidate
-                ]
-            except Exception:
-                raise ValueError(
-                    f"{parameter_to_invalidate} not found in provided parameters."
-                ) from None
-            values_from_constraint = None
-        else:
-            parameter_to_invalidate = choice(
-                tuple(relation_property_names.intersection(parameter_names))
-            )
-            try:
-                [parameter_data] = [
-                    d
+        # ensure we're not modifying mutable properties
+        params = deepcopy(request_data.params)
+        headers = deepcopy(request_data.headers)
+
+        if status_code == self.invalid_property_default_response:
+            parameter_names = set(params.keys()).union(set(headers.keys()))
+            parameter_names.update(relation_property_names)
+            # if all parameters are optional and none were provided, randomly pick one
+            if not parameter_names:
+                parameter_names = {
+                    d["name"]
                     for d in request_data.parameters
-                    if d["name"] == parameter_to_invalidate
-                ]
-            except Exception:
-                raise ValueError(
-                    f"{parameter_to_invalidate} from Relation not found in provided parameters."
-                ) from None
-            relations_for_parameter = [
+                    if d["in"] in ["query", "header"]
+                }
+        else:
+            # non-default status_codes can only be the result of a Relation
+            parameter_names = relation_property_names
+
+        parameter_to_invalidate = choice(tuple(parameter_names))
+        # check for invalid parameters in the provided request_data
+        try:
+            [parameter_data] = [
+                data
+                for data in request_data.parameters
+                if data["name"] == parameter_to_invalidate
+            ]
+        except Exception:
+            raise ValueError(
+                f"{parameter_to_invalidate} not found in provided parameters."
+            ) from None
+        # get the constraint values if available for the chosen parameter
+        try:
+            [values_from_constraint] = [
                 r.values
                 for r in relations_for_status_code
                 if r.property_name == parameter_to_invalidate
             ]
-            values_from_constraint = relations_for_parameter[0]
+        except ValueError:
+            values_from_constraint = []
+        # if the parameter was not provided, add it to params / headers
+        params, headers = self.ensure_parameter_in_parameters(
+            parameter_to_invalidate=parameter_to_invalidate,
+            params=params,
+            headers=headers,
+            parameter_data=parameter_data,
+            values_from_constraint=values_from_constraint,
+        )
+        # determine the invalid_value
         if parameter_to_invalidate in params.keys():
             valid_value = params[parameter_to_invalidate]
         else:
@@ -753,13 +921,45 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
             current_value=valid_value,
             values_from_constraint=values_from_constraint,
         )
-
+        # update the params / headers and return
         if parameter_to_invalidate in params.keys():
             params[parameter_to_invalidate] = invalid_value
         else:
             headers[parameter_to_invalidate] = str(invalid_value)
         return params, headers
 
+    @staticmethod
+    def ensure_parameter_in_parameters(
+        parameter_to_invalidate: str,
+        params: Dict[str, Any],
+        headers: Dict[str, str],
+        parameter_data: Dict[str, Any],
+        values_from_constraint: List[Any],
+    ):
+        """
+        Returns the params, headers tuple with parameter_to_invalidate with a valid
+        value to params or headers if not originally present.
+        """
+        if (
+            parameter_to_invalidate not in params.keys()
+            and parameter_to_invalidate not in headers.keys()
+        ):
+            if values_from_constraint:
+                valid_value = choice(values_from_constraint)
+            else:
+                valid_value = value_utils.get_valid_value(parameter_data["schema"])
+            if (
+                parameter_data["in"] == "query"
+                and parameter_to_invalidate not in params.keys()
+            ):
+                params[parameter_to_invalidate] = valid_value
+            if (
+                parameter_data["in"] == "header"
+                and parameter_to_invalidate not in headers.keys()
+            ):
+                headers[parameter_to_invalidate] = valid_value
+        return params, headers
+
     @keyword
     def ensure_in_use(self, url: str, resource_relation: IdReference) -> None:
         """
@@ -783,10 +983,7 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
         request_data = self.get_request_data(
             method="post", endpoint=resource_relation.post_path
         )
-        params = request_data.params
-        headers = request_data.headers
-        dto = request_data.dto
-        json_data = asdict(dto)
+        json_data = asdict(request_data.dto)
         json_data[resource_relation.property_name] = resource_id
         post_url: str = run_keyword(
             "get_valid_url",
@@ -794,7 +991,12 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
             "post",
         )
         response: Response = run_keyword(
-            "authorized_request", post_url, "post", params, headers, json_data
+            "authorized_request",
+            post_url,
+            "post",
+            request_data.params,
+            request_data.headers,
+            json_data,
         )
         if not response.ok:
             logger.debug(
@@ -855,7 +1057,7 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
         )
 
     @keyword
-    def authorized_request(
+    def authorized_request(  # pylint: disable=too-many-arguments
         self,
         url: str,
         method: str,
